﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Diagnostics;

namespace Maze {
	class Program {
		static int wallWidth = 3;
		static int wallHeight = 2;

		static int roomWidth = 4;
		static int roomHeight = 2;

		static int mapWidth;
		static int mapHeight;
		static char[,] baseMap;

		private static void initialize() {
			mapWidth = (roomWidth + 1) * (wallWidth + 1) + 1;
			mapHeight = (roomHeight + 1) * (wallHeight + 1) + 1;
			baseMap = new char[mapHeight, mapWidth];

			int numRows = roomHeight;
			int numCols = roomWidth;
			int midRow = (int)Math.Ceiling(numRows / 2.0);
			int midCol = (int)Math.Ceiling(numCols / 2.0);

			for (int i = 0; i < mapHeight; i++) {
				if (i <= midRow * (wallHeight+1) || i >= (midRow + 1) * (wallHeight+1)) {
					baseMap[i, 0] = 'X';
					baseMap[i, mapWidth - 1] = 'X';
				}
			}
			for (int j = 0; j < mapWidth; j++) {
				if (j <= midCol * (wallWidth+1) || j >= (midCol + 1) * (wallWidth+1)) {
					baseMap[0, j] = 'X';
					baseMap[mapHeight - 1, j] = 'X';
				}
			}
			for (int i = 0; i < roomHeight; i++) {
				for (int j = 0; j < roomWidth; j++) {
					int mapI = (i + 1) * (wallHeight + 1);
					int mapJ = (j + 1) * (wallWidth + 1);
					baseMap[mapI, mapJ] = 'O';
				}
			}
		}

		static void Main(string[] args) {
			initialize();

			bool playing = true;
			Maze maze = null;
			bool newMap = true;
			char[,] map = null;
			int playerI = 1;
			int playerJ = 1;
			createMap(ref maze, ref newMap, ref playerI, ref playerJ);

			Console.CursorVisible = false;
			while (playing) {
				if (newMap) {
					map = drawMap(maze);
					newMap = false;
				}

				drawPlayer(playerI, playerJ);
				var key = Console.ReadKey(true);
				drawPlayer(playerI, playerJ, true);

				movePlayer(map, ref playerI, ref playerJ, key);
				newMap = moveMap(maze, ref playerI, ref playerJ);
				handleSpecialKeys(ref maze, ref newMap, ref playerI, ref playerJ, ref key, ref map, ref playing);
			}
		}

		public static void move(Maze maze, Direction direction) {
			int roomNumber = maze.RoomNumber;
			
			switch (direction) {
				case Direction.North:
					roomNumber += 10000;
					break;
				case Direction.South:
					roomNumber -= 10000;
					break;
				case Direction.East:
					roomNumber += 1;
					break;
				case Direction.West:
					roomNumber -= 1;
					break;
				default:
					break;
			}

			maze.RoomNumber = roomNumber;
		}

		private static void handleSpecialKeys(ref Maze maze, ref bool newMap, ref int playerI, ref int playerJ, ref ConsoleKeyInfo key, ref char[,] map, ref bool playing) {
			if (key.Key == ConsoleKey.R) {
				createMap(ref maze, ref newMap, ref playerI, ref playerJ);
			} else if (key.Key == ConsoleKey.S) {
				newMap = true;
				int room;
				Console.SetCursorPosition(0, mapHeight + 2);
				if (!input("Jump to which room? ", out room)) { return; }
				maze.GenerateRoom(room);
				playerI = 1;
				playerJ = 1;
			} else if (key.Key == ConsoleKey.D) {
				newMap = true;
				int width, height;
				Console.SetCursorPosition(0, mapHeight + 2);
				if (!input(string.Format("Room width ({0}): ", roomWidth), out width)) { return; }
				if (!input(string.Format("Room height ({0}): ", roomHeight), out height)) { return; }
				roomWidth = width;
				roomHeight = height;
				initialize();
				createMap(ref maze, ref newMap, ref playerI, ref playerJ);
			} else if (key.Key == ConsoleKey.H) {
				newMap = true;
				Console.Clear();
				Console.ForegroundColor = ConsoleColor.White;
				Console.WriteLine("Arrow keys to move.");
				Console.WriteLine("'Esc' to quit.");
				Console.WriteLine("'h' for this help screen.");
				Console.WriteLine("'r' to re-seed the map.");
				Console.WriteLine("'s' to set the room number.");
				Console.WriteLine("'d' to set room dimensions.");
				Console.WriteLine("'l' to set wall size.");
				Console.WriteLine("<Press any key to continue>");
				Console.ReadKey(true);
			} else if (key.Key == ConsoleKey.Escape) {
				playing = false;
			} else if (key.Key == ConsoleKey.L) {
				newMap = true;
				int width, height;
				Console.SetCursorPosition(0, mapHeight + 2);
				if (!input(string.Format("Wall width ({0}): ", wallWidth), out width)) { return; }
				if (!input(string.Format("Wall height ({0}): ", wallHeight), out height)) { return; }
				wallWidth = width;
				wallHeight = height;
				initialize();
				//createMap(ref maze, ref newMap, ref playerI, ref playerJ);
			}
		}

		private static void createMap(ref Maze maze, ref bool newMap, ref int playerI, ref int playerJ) {
			maze = new Maze(roomWidth, roomHeight, 4);
			newMap = true;
			playerI = 1;
			playerJ = 1;
		}

		private static bool input(string prompt, out int value) {
			Console.Write(prompt);
			string input = Console.ReadLine();
			if (!int.TryParse(input, out value)) {
				Console.WriteLine("Invalid input.");
				Thread.Sleep(1000);
				return false;
			}
			return true;
		}

		private static void drawPlayer(int playerI, int playerJ, bool clear = false) {
			Console.ForegroundColor = ConsoleColor.Cyan;
			Console.SetCursorPosition(playerJ, playerI + 2);
			Console.Write(clear ? ' ' : '@');
		}

		private static void movePlayer(char[,] map, ref int playerI, ref int playerJ, ConsoleKeyInfo key) {
			int backupI = playerI;
			int backupJ = playerJ;

			if (key.Key == ConsoleKey.LeftArrow) {
				playerJ -= 1;
			} else if (key.Key == ConsoleKey.RightArrow) {
				playerJ += 1;
			} else if (key.Key == ConsoleKey.DownArrow) {
				playerI += 1;
			} else if (key.Key == ConsoleKey.UpArrow) {
				playerI -= 1;
			}

			if (map[playerI, playerJ] != '\0') {
				playerI = backupI;
				playerJ = backupJ;
			}
		}

		private static bool moveMap(Maze maze, ref int playerI, ref int playerJ) {
			bool newMap = false;
			if (playerI == 0) {
				move(maze, Direction.North);
				newMap = true;
				playerI = mapHeight - 2;
			} else if (playerI == mapHeight - 1) {
				move(maze, Direction.South);
				newMap = true;
				playerI = 1;
			} else if (playerJ == 0) {
				move(maze, Direction.West);
				newMap = true;
				playerJ = mapWidth - 2;
			} else if (playerJ == mapWidth - 1) {
				move(maze, Direction.East);
				newMap = true;
				playerJ = 1;
			}
			return newMap;
		}

		private static char[,] drawMap(Maze maze) {
			Console.Clear();
			Console.ForegroundColor = ConsoleColor.White;
			Console.WriteLine("Arrow keys to move; 'Esc' to quit; 'h' for help.");
			Console.ForegroundColor = ConsoleColor.Yellow;
			Console.WriteLine("Room number: {0}", maze.RoomNumber);
			int[,] room = maze.GenerateRoom();
			char[,] map = generateMap(room);
			for (int i = 0; i < mapHeight; i++) {
				for (int j = 0; j < mapWidth; j++) {
					char tile = map[i, j];
					switch (tile) {
						case 'X':
							Console.ForegroundColor = ConsoleColor.Blue;
							break;
						case 'O':
							Console.ForegroundColor = ConsoleColor.DarkGreen;
							break;
						case '-':
						case '|':
							Console.ForegroundColor = ConsoleColor.Red;
							break;
						case '\0':
							tile = ' ';
							break;
					}
					Console.Write(tile);
				}
				Console.WriteLine();
			}
			return map;
		}

		private static char[,] generateMap(int[,] room) {
			char[,] map = (char[,])baseMap.Clone();

			for (int i = 0; i < roomHeight; i++) {
				for (int j = 0; j < roomWidth; j++) {
					Direction dir = (Direction)room[i, j];
					int mapI = (i + 1) * (wallHeight + 1);
					int mapJ = (j + 1) * (wallWidth + 1);
					switch (dir) {
						case Direction.North:
							for (int length = 1; length <= wallHeight; length++) {
								map[mapI - length, mapJ] = '|';
							}
							break;
						case Direction.East:
							for (int length = 1; length <= wallWidth; length++) {
								map[mapI, mapJ + length] = '-';
							}
							break;
						case Direction.South:
							for (int length = 1; length <= wallHeight; length++) {
								map[mapI + length, mapJ] = '|';
							}
							break;
						case Direction.West:
							for (int length = 1; length <= wallWidth; length++) {
								map[mapI, mapJ - length] = '-';
							}
							break;
						default:
							break;
					}
				}
			}

			return map;
		}
	}
}
